@@ -1,3 +1,6 @@
+// Copyright (c) The Thanos Authors.
+// Licensed under the Apache License 2.0.
+
 package store
 
 import (
@@ -12,22 +15,27 @@ import (
 	"net/url"
 	"path"
 	"sort"
+	"strconv"
 	"strings"
 	"sync"
+	"time"
 
 	"github.com/go-kit/kit/log"
 	"github.com/go-kit/kit/log/level"
 	"github.com/gogo/protobuf/proto"
 	"github.com/golang/snappy"
-	opentracing "github.com/opentracing/opentracing-go"
+	"github.com/opentracing/opentracing-go"
 	"github.com/pkg/errors"
-	"github.com/prometheus/prometheus/prompb"
+	"github.com/prometheus/common/version"
+	"github.com/prometheus/prometheus/pkg/labels"
+	"github.com/prometheus/prometheus/pkg/timestamp"
 	"github.com/prometheus/prometheus/storage/remote"
 	"github.com/prometheus/prometheus/tsdb/chunkenc"
-	"github.com/prometheus/prometheus/tsdb/labels"
 	"github.com/thanos-io/thanos/pkg/component"
+	"github.com/thanos-io/thanos/pkg/exthttp"
 	"github.com/thanos-io/thanos/pkg/runutil"
 	"github.com/thanos-io/thanos/pkg/store/storepb"
+	"github.com/thanos-io/thanos/pkg/store/storepb/prompb"
 	"github.com/thanos-io/thanos/pkg/tracing"
 	"google.golang.org/grpc/codes"
 	"google.golang.org/grpc/status"
@@ -41,6 +49,8 @@ var statusToCode = map[int]codes.Code{
 	http.StatusInternalServerError: codes.Internal,
 }
 
+var userAgent = fmt.Sprintf("Thanos/%s", version.Version)
+
 // PrometheusStore implements the store node API on top of the Prometheus remote read API.
 type PrometheusStore struct {
 	logger         log.Logger
@@ -54,6 +64,12 @@ type PrometheusStore struct {
 	remoteReadAcceptableResponses []prompb.ReadRequest_ResponseType
 }
 
+const (
+	initialBufSize = 32 * 1024 // 32KB seems like a good minimum starting size.
+
+	SUCCESS = "success"
+)
+
 // NewPrometheusStore returns a new PrometheusStore that uses the given HTTP client
 // to talk to Prometheus.
 // It attaches the provided external labels to all results.
@@ -70,7 +86,7 @@ func NewPrometheusStore(
 	}
 	if client == nil {
 		client = &http.Client{
-			Transport: tracing.HTTPTripperware(logger, http.DefaultTransport),
+			Transport: tracing.HTTPTripperware(logger, exthttp.NewTransport()),
 		}
 	}
 	p := &PrometheusStore{
@@ -81,6 +97,10 @@ func NewPrometheusStore(
 		externalLabels:                externalLabels,
 		timestamps:                    timestamps,
 		remoteReadAcceptableResponses: []prompb.ReadRequest_ResponseType{prompb.ReadRequest_STREAMED_XOR_CHUNKS, prompb.ReadRequest_SAMPLES},
+		buffers: sync.Pool{New: func() interface{} {
+			b := make([]byte, 0, initialBufSize)
+			return &b
+		}},
 	}
 	return p, nil
 }
@@ -118,10 +138,6 @@ func (p *PrometheusStore) Info(ctx context.Context, r *storepb.InfoRequest) (*st
 
 func (p *PrometheusStore) getBuffer() *[]byte {
 	b := p.buffers.Get()
-	if b == nil {
-		buf := make([]byte, 0, 32*1024) // 32KB seems like a good minimum starting size.
-		return &buf
-	}
 	return b.(*[]byte)
 }
 
@@ -143,7 +159,7 @@ func (p *PrometheusStore) Series(r *storepb.SeriesRequest, s storepb.Store_Serie
 	}
 
 	if len(newMatchers) == 0 {
-		return status.Error(codes.InvalidArgument, errors.New("no matchers specified (excluding external labels)").Error())
+		return status.Error(codes.InvalidArgument, "no matchers specified (excluding external labels)")
 	}
 
 	// Don't ask for more than available time. This includes potential `minTime` flag limit.
@@ -152,6 +168,27 @@ func (p *PrometheusStore) Series(r *storepb.SeriesRequest, s storepb.Store_Serie
 		r.MinTime = availableMinTime
 	}
 
+	if r.SkipChunks {
+		labelMaps, err := p.seriesLabels(s.Context(), newMatchers, r.MinTime, r.MaxTime)
+		if err != nil {
+			return err
+		}
+		for _, lbm := range labelMaps {
+			lset := make([]storepb.Label, 0, len(lbm)+len(externalLabels))
+			for k, v := range lbm {
+				lset = append(lset, storepb.Label{Name: k, Value: v})
+			}
+			lset = append(lset, storepb.PromLabelsToLabelsUnsafe(externalLabels)...)
+			sort.Slice(lset, func(i, j int) bool {
+				return lset[i].Name < lset[j].Name
+			})
+			if err = s.Send(storepb.NewSeriesResponse(&storepb.Series{Labels: lset})); err != nil {
+				return err
+			}
+		}
+		return nil
+	}
+
 	q := &prompb.Query{StartTimestampMs: r.MinTime, EndTimestampMs: r.MaxTime}
 
 	for _, m := range newMatchers {
@@ -324,7 +361,7 @@ func (p *PrometheusStore) handleStreamedPrometheusResponse(s storepb.Store_Serie
 	return nil
 }
 
-func (p *PrometheusStore) fetchSampledResponse(ctx context.Context, resp *http.Response) (*prompb.ReadResponse, error) {
+func (p *PrometheusStore) fetchSampledResponse(ctx context.Context, resp *http.Response) (_ *prompb.ReadResponse, err error) {
 	defer runutil.ExhaustCloseWithLogOnErr(p.logger, resp.Body, "prom series request body")
 
 	b := p.getBuffer()
@@ -333,21 +370,24 @@ func (p *PrometheusStore) fetchSampledResponse(ctx context.Context, resp *http.R
 	if _, err := io.Copy(buf, resp.Body); err != nil {
 		return nil, errors.Wrap(err, "copy response")
 	}
-	spanSnappyDecode, ctx := tracing.StartSpan(ctx, "decompress_response")
+
 	sb := p.getBuffer()
-	decomp, err := snappy.Decode(*sb, buf.Bytes())
-	spanSnappyDecode.Finish()
+	var decomp []byte
+	tracing.DoInSpan(ctx, "decompress_response", func(ctx context.Context) {
+		decomp, err = snappy.Decode(*sb, buf.Bytes())
+	})
 	defer p.putBuffer(sb)
 	if err != nil {
 		return nil, errors.Wrap(err, "decompress response")
 	}
 
 	var data prompb.ReadResponse
-	spanUnmarshal, _ := tracing.StartSpan(ctx, "unmarshal_response")
-	if err := proto.Unmarshal(decomp, &data); err != nil {
+	tracing.DoInSpan(ctx, "unmarshal_response", func(ctx context.Context) {
+		err = proto.Unmarshal(decomp, &data)
+	})
+	if err != nil {
 		return nil, errors.Wrap(err, "unmarshal response")
 	}
-	spanUnmarshal.Finish()
 	if len(data.Results) != 1 {
 		return nil, errors.Errorf("unexpected result size %d", len(data.Results))
 	}
@@ -381,7 +421,7 @@ func (p *PrometheusStore) chunkSamples(series *prompb.TimeSeries, maxSamplesPerC
 	return chks, nil
 }
 
-func (p *PrometheusStore) startPromSeries(ctx context.Context, q *prompb.Query) (*http.Response, error) {
+func (p *PrometheusStore) startPromSeries(ctx context.Context, q *prompb.Query) (presp *http.Response, err error) {
 	reqb, err := proto.Marshal(&prompb.ReadRequest{
 		Queries:               []*prompb.Query{q},
 		AcceptedResponseTypes: p.remoteReadAcceptableResponses,
@@ -390,6 +430,11 @@ func (p *PrometheusStore) startPromSeries(ctx context.Context, q *prompb.Query)
 		return nil, errors.Wrap(err, "marshal read request")
 	}
 
+	qjson, err := json.Marshal(q)
+	if err != nil {
+		return nil, errors.Wrap(err, "json encode query for tracing")
+	}
+
 	u := *p.base
 	u.Path = path.Join(u.Path, "api/v1/read")
 
@@ -399,13 +444,14 @@ func (p *PrometheusStore) startPromSeries(ctx context.Context, q *prompb.Query)
 	}
 	preq.Header.Add("Content-Encoding", "snappy")
 	preq.Header.Set("Content-Type", "application/x-stream-protobuf")
-	spanReqDo, ctx := tracing.StartSpan(ctx, "query_prometheus_request")
-	preq = preq.WithContext(ctx)
-	presp, err := p.client.Do(preq)
+	preq.Header.Set("User-Agent", userAgent)
+	tracing.DoInSpan(ctx, "query_prometheus_request", func(ctx context.Context) {
+		preq = preq.WithContext(ctx)
+		presp, err = p.client.Do(preq)
+	}, opentracing.Tag{Key: "prometheus.query", Value: string(qjson)})
 	if err != nil {
 		return nil, errors.Wrap(err, "send request")
 	}
-	spanReqDo.Finish()
 	if presp.StatusCode/100 != 2 {
 		// Best effort read.
 		b, err := ioutil.ReadAll(presp.Body)
@@ -419,7 +465,7 @@ func (p *PrometheusStore) startPromSeries(ctx context.Context, q *prompb.Query)
 	return presp, nil
 }
 
-// matchesExternalLabels filters out external labels matching from matcher if exsits as the local storage does not have them.
+// matchesExternalLabels filters out external labels matching from matcher if exists as the local storage does not have them.
 // It also returns false if given matchers are not matching external labels.
 func matchesExternalLabels(ms []storepb.LabelMatcher, externalLabels labels.Labels) (bool, []storepb.LabelMatcher, error) {
 	if len(externalLabels) == 0 {
@@ -466,33 +512,35 @@ func (p *PrometheusStore) encodeChunk(ss []prompb.Sample) (storepb.Chunk_Encodin
 }
 
 // translateAndExtendLabels transforms a metrics into a protobuf label set. It additionally
-// attaches the given labels to it, overwriting existing ones on colllision.
+// attaches the given labels to it, overwriting existing ones on collision.
+// Both input labels are expected to be sorted.
+//
+// NOTE(bwplotka): Don't use modify passed slices as we reuse underlying memory.
 func (p *PrometheusStore) translateAndExtendLabels(m []prompb.Label, extend labels.Labels) []storepb.Label {
+	pbLabels := storepb.PrompbLabelsToLabelsUnsafe(m)
+	pbExtend := storepb.PromLabelsToLabelsUnsafe(extend)
+
 	lset := make([]storepb.Label, 0, len(m)+len(extend))
+	ei := 0
 
-	for _, l := range m {
-		if extend.Get(l.Name) != "" {
-			continue
+Outer:
+	for _, l := range pbLabels {
+		for ei < len(pbExtend) {
+			if l.Name < pbExtend[ei].Name {
+				break
+			}
+			lset = append(lset, pbExtend[ei])
+			ei++
+			if l.Name == pbExtend[ei-1].Name {
+				continue Outer
+			}
 		}
-		lset = append(lset, storepb.Label{
-			Name:  l.Name,
-			Value: l.Value,
-		})
+		lset = append(lset, l)
 	}
-
-	return extendLset(lset, extend)
-}
-
-func extendLset(lset []storepb.Label, extend labels.Labels) []storepb.Label {
-	for _, l := range extend {
-		lset = append(lset, storepb.Label{
-			Name:  l.Name,
-			Value: l.Value,
-		})
+	for ei < len(pbExtend) {
+		lset = append(lset, pbExtend[ei])
+		ei++
 	}
-	sort.Slice(lset, func(i, j int) bool {
-		return lset[i].Name < lset[j].Name
-	})
 	return lset
 }
 
@@ -507,6 +555,7 @@ func (p *PrometheusStore) LabelNames(ctx context.Context, _ *storepb.LabelNamesR
 	if err != nil {
 		return nil, status.Error(codes.Internal, err.Error())
 	}
+	req.Header.Set("User-Agent", userAgent)
 
 	span, ctx := tracing.StartSpan(ctx, "/prom_label_names HTTP[client]")
 	defer span.Finish()
@@ -518,7 +567,7 @@ func (p *PrometheusStore) LabelNames(ctx context.Context, _ *storepb.LabelNamesR
 	defer runutil.ExhaustCloseWithLogOnErr(p.logger, resp.Body, "label names request body")
 
 	if resp.StatusCode/100 != 2 {
-		return nil, status.Error(codes.Internal, fmt.Sprintf("request Prometheus server failed, code %s", resp.Status))
+		return nil, status.Errorf(codes.Internal, "request Prometheus server failed, code %s", resp.Status)
 	}
 
 	if resp.StatusCode == http.StatusNoContent {
@@ -540,7 +589,7 @@ func (p *PrometheusStore) LabelNames(ctx context.Context, _ *storepb.LabelNamesR
 		return nil, status.Error(codes.Internal, err.Error())
 	}
 
-	if m.Status != "success" {
+	if m.Status != SUCCESS {
 		code, exists := statusToCode[resp.StatusCode]
 		if !exists {
 			return nil, status.Error(codes.Internal, m.Error)
@@ -567,6 +616,7 @@ func (p *PrometheusStore) LabelValues(ctx context.Context, r *storepb.LabelValue
 	if err != nil {
 		return nil, status.Error(codes.Internal, err.Error())
 	}
+	req.Header.Set("User-Agent", userAgent)
 
 	span, ctx := tracing.StartSpan(ctx, "/prom_label_values HTTP[client]")
 	defer span.Finish()
@@ -578,7 +628,7 @@ func (p *PrometheusStore) LabelValues(ctx context.Context, r *storepb.LabelValue
 	defer runutil.ExhaustCloseWithLogOnErr(p.logger, resp.Body, "label values request body")
 
 	if resp.StatusCode/100 != 2 {
-		return nil, status.Error(codes.Internal, fmt.Sprintf("request Prometheus server failed, code %s", resp.Status))
+		return nil, status.Errorf(codes.Internal, "request Prometheus server failed, code %s", resp.Status)
 	}
 
 	if resp.StatusCode == http.StatusNoContent {
@@ -601,7 +651,7 @@ func (p *PrometheusStore) LabelValues(ctx context.Context, r *storepb.LabelValue
 
 	sort.Strings(m.Data)
 
-	if m.Status != "success" {
+	if m.Status != SUCCESS {
 		code, exists := statusToCode[resp.StatusCode]
 		if !exists {
 			return nil, status.Error(codes.Internal, m.Error)
@@ -611,3 +661,72 @@ func (p *PrometheusStore) LabelValues(ctx context.Context, r *storepb.LabelValue
 
 	return &storepb.LabelValuesResponse{Values: m.Data}, nil
 }
+
+// seriesLabels returns the labels from Prometheus series API.
+func (p *PrometheusStore) seriesLabels(ctx context.Context, matchers []storepb.LabelMatcher, startTime, endTime int64) ([]map[string]string, error) {
+	u := *p.base
+	u.Path = path.Join(u.Path, "/api/v1/series")
+	q := u.Query()
+
+	metric, err := matchersToString(matchers)
+	if err != nil {
+		return nil, errors.Wrap(err, "invalid matchers")
+	}
+
+	q.Add("match[]", metric)
+	q.Add("start", formatTime(timestamp.Time(startTime)))
+	q.Add("end", formatTime(timestamp.Time(endTime)))
+	u.RawQuery = q.Encode()
+
+	req, err := http.NewRequest("GET", u.String(), nil)
+	if err != nil {
+		return nil, status.Error(codes.Internal, err.Error())
+	}
+	req.Header.Set("User-Agent", userAgent)
+
+	span, ctx := tracing.StartSpan(ctx, "/prom_series HTTP[client]")
+	defer span.Finish()
+
+	resp, err := p.client.Do(req.WithContext(ctx))
+	if err != nil {
+		return nil, status.Error(codes.Internal, err.Error())
+	}
+	defer runutil.ExhaustCloseWithLogOnErr(p.logger, resp.Body, "series request body")
+
+	if resp.StatusCode/100 != 2 {
+		return nil, status.Errorf(codes.Internal, "request Prometheus server failed, code %s", resp.Status)
+	}
+
+	if resp.StatusCode == http.StatusNoContent {
+		return nil, nil
+	}
+
+	var m struct {
+		Data   []map[string]string `json:"data"`
+		Status string              `json:"status"`
+		Error  string              `json:"error"`
+	}
+
+	body, err := ioutil.ReadAll(resp.Body)
+	if err != nil {
+		return nil, status.Error(codes.Internal, err.Error())
+	}
+
+	if err = json.Unmarshal(body, &m); err != nil {
+		return nil, status.Error(codes.Internal, err.Error())
+	}
+
+	if m.Status != SUCCESS {
+		code, exists := statusToCode[resp.StatusCode]
+		if !exists {
+			return nil, status.Error(codes.Internal, m.Error)
+		}
+		return nil, status.Error(code, m.Error)
+	}
+
+	return m.Data, nil
+}
+
+func formatTime(t time.Time) string {
+	return strconv.FormatFloat(float64(t.Unix())+float64(t.Nanosecond())/1e9, 'f', -1, 64)
+}
